@@ -12,7 +12,7 @@ import numpy as np
 from aim import Image
 import matplotlib.pyplot as plt
 from typing import List, Union
-from .utils import QuantileLossMO
+from .utils import QuantileLossMO, CPRS
 import torch.nn as nn
 
 
@@ -137,10 +137,15 @@ class Base(pl.LightningModule):
         if n_classes==0:
             self.is_classification = False
             if len(self.quantiles)>0:
-                assert len(self.quantiles)==3, beauty_string('ONLY 3 quantiles premitted','info',True)
-                self.use_quantiles = True
-                self.mul = len(self.quantiles)
-                self.loss = QuantileLossMO(quantiles)
+                if self.loss_type=='cprs':
+                    self.use_quantiles = True
+                    self.mul = len(self.quantiles)
+                    self.loss = CPRS(alpha=self.persistence_weight)
+                else:
+                    assert len(self.quantiles)==3, beauty_string('ONLY 3 quantiles premitted','info',True)
+                    self.use_quantiles = True
+                    self.mul = len(self.quantiles)
+                    self.loss = QuantileLossMO(quantiles)
             else:
                 self.use_quantiles = False
                 self.mul = 1
@@ -157,7 +162,7 @@ class Base(pl.LightningModule):
 
 
         self.future_steps = future_steps
-        
+        self.return_additional_loss = False
         beauty_string(self.description,'info',True)
     @abstractmethod
     def forward(self, batch:dict)-> torch.tensor:
@@ -189,6 +194,13 @@ class Base(pl.LightningModule):
         Returns:
             torch.tensor: result
         """
+        
+        if self.loss_type=='cprs':
+            tmp = self(batch)
+            tmp = torch.quantile(tmp, torch.tensor([0.05, 0.5, 0.95]), dim=-1).permute(1,2,3,0)
+            return tmp
+            #return tmp.mean(axis=-1).unsqueeze(-1)
+        
         return self(batch)
         
     def configure_optimizers(self):
@@ -250,14 +262,22 @@ class Base(pl.LightningModule):
             opt = self.optimizers()
             def closure():
                 opt.zero_grad()
-                y_hat = self(batch)
-                loss = self.compute_loss(batch,y_hat)
+                if self.return_additional_loss:
+                    y_hat,score = self(batch)
+                    loss = self.compute_loss(batch,y_hat) + score
+                else:
+                    y_hat = self(batch)
+                    loss = self.compute_loss(batch,y_hat)
                 self.manual_backward(loss)
                 return loss
 
             opt.step(closure)
-            y_hat = self(batch)
-            loss = self.compute_loss(batch,y_hat)
+            if self.return_additional_loss:
+                y_hat,score = self(batch)
+                loss = self.compute_loss(batch,y_hat)+score
+            else:
+                y_hat = self(batch)
+                loss = self.compute_loss(batch,y_hat)
             
             #opt.first_step(zero_grad=True)
 
@@ -272,8 +292,12 @@ class Base(pl.LightningModule):
    
             #self.trainer.fit_loop.epoch_loop.manual_optimization.optim_step_progress.increment("optimizer")
         else:
-            y_hat = self(batch)
-            loss = self.compute_loss(batch,y_hat)
+            if self.return_additional_loss:
+                y_hat,score = self(batch)
+                loss = self.compute_loss(batch,y_hat)+score
+            else:
+                y_hat = self(batch)
+                loss = self.compute_loss(batch,y_hat)
             
         self.train_epoch_metrics.append(loss.item())
         return loss
@@ -285,7 +309,12 @@ class Base(pl.LightningModule):
         
         :meta private:
         """
-        y_hat = self(batch)
+
+        if self.return_additional_loss:
+            y_hat,score = self(batch)
+        else:
+            y_hat = self(batch)
+            score = 0
         if batch_idx==0:
             if self.use_quantiles:
                 idx = 1
@@ -305,7 +334,7 @@ class Base(pl.LightningModule):
                     ax.set_title(f'Channel {i} first element first batch validation {int(100*self.count_epoch/self.trainer.max_epochs)}%')
                     self.logger.experiment.track(Image(fig), name='cm_training_end')
                     #self.log(f"example_{i}", np.stack([real, pred]).T,sync_dist=True)
-        self.validation_epoch_metrics.append(self.compute_loss(batch,y_hat))
+        self.validation_epoch_metrics.append(self.compute_loss(batch,y_hat)+score)
         return 
 
 
@@ -315,7 +344,12 @@ class Base(pl.LightningModule):
         
         :meta private:
         """
-        avg = torch.stack(self.validation_epoch_metrics).mean()
+
+        if len(self.validation_epoch_metrics)==0:
+            avg = 10000
+            beauty_string(f'THIS IS A BUG, It should be polulated','info',self.verbose)
+        else:
+            avg = torch.stack(self.validation_epoch_metrics).mean()
         self.validation_epoch_metrics = []
         self.log("val_loss", avg,sync_dist=True)
         beauty_string(f'Epoch: {self.count_epoch} train error: {self.train_loss_epoch:.4f} validation loss: {avg:.4f}','info',self.verbose)
@@ -327,7 +361,11 @@ class Base(pl.LightningModule):
         
         :meta private:
         """
-        avg = np.stack(self.train_epoch_metrics).mean()
+        if len(self.train_epoch_metrics)==0:
+            avg = 0
+            beauty_string(f'THIS IS A BUG, It should be polulated','info',self.verbose)
+        else:
+            avg = np.stack(self.train_epoch_metrics).mean()
         self.log("train_loss", avg,sync_dist=True)
         self.count_epoch+=1    
         self.train_epoch_metrics = []
@@ -339,6 +377,16 @@ class Base(pl.LightningModule):
         
         :meta private:
         """
+        if self.loss_type=='cprs':
+            return self.loss(y_hat,batch['y'])
+
+        if self.loss_type=='long_lag':
+
+            batch_size,width,n_variables = batch['y'].shape
+            tmp = torch.abs(y_hat[:,:,:,0]-batch['y'])*torch.linspace(1,self.persistence_weight,width).view(1,width,1).repeat(batch_size,1,n_variables)
+            return tmp.mean()
+
+
 
         if self.use_quantiles is False:
             initial_loss = self.loss(y_hat[:,:,:,0], batch['y'])
@@ -411,24 +459,16 @@ class Base(pl.LightningModule):
             
         elif self.loss_type=='dilated':
             #BxLxCxMUL
-            if self.persistence_weight==0.1:
-                alpha = 0.25
-            if self.persistence_weight==1:
-                alpha = 0.5
-            else:
-                alpha  =0.75
+
             alpha = self.persistence_weight 
             gamma = 0.01
             loss = 0
             ##no multichannel here
-            for i in range(y_hat.shape[2]):
-                ##error here
-                
+            for i in range(y_hat.shape[2]):                
                 loss+= dilate_loss( batch['y'][:,:,i:i+1],x[:,:,i:i+1], alpha, gamma, y_hat.device)
             
         elif self.loss_type=='huber':
-            loss = torch.nn.HuberLoss(reduction='mean', delta=self.persistence_weight/10)   
-            #loss = torch.nn.HuberLoss(reduction='mean', delta=self.persistence_weight)   
+            loss = torch.nn.HuberLoss(reduction='mean', delta=self.persistence_weight)   
             if self.use_quantiles is False:
                 x = y_hat[:,:,:,0]
             else:
